"""
Name: GALE_BatchPerformance.py
Author: Ryan Urbanowicz - written at Dartmouth College, Hanover, NH, USA 
Note: code primarily based on Java GALE authored by Xavier Llor <xllora@illigal.ge.uiuc.edu>
Contact: ryan.j.urbanowicz@darmouth.edu
Created: 2/22/2010
Updated: 10/15/2013
Description: Creates a performance tracking object
---------------------------------------------------------------------------------------------------------------------------------------------------------
GALE Version 1.0: A Python implementation of the GALE learning classifier system algorithm implemented for SNP genetic data mining and classification tasks.
Copyright (C) 2013 Ryan Urbanowicz 
This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the 
Free Software Foundation; either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABLILITY 
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, 
Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
---------------------------------------------------------------------------------------------------------------------------------------------------------
"""

#!/usr/bin/env python

#Import modules
from GALE_Constants import * 
import math

class GALEBatchPerformance:
    def __init__(self, nClass): 
        """ Initialize performance objects for a single agent. """
        self.iConfusionMatrix = []   #Confusion Matrix for tracking performance
        self.iDontKnow = 0  #Number of unclassified instances - (no match i believe)
        self.iTotal = 0     #Number of instances
        self.numClass = nClass  #Number of classes
        
        # Construct Confusion Matrix
        for i in range(self.numClass):
            self.iConfusionMatrix.append([])
            for j in range(self.numClass):
                self.iConfusionMatrix[i].append(0)
   
   
    def getConfusionMatrix(self):
        """ Return the confusion matrix. """
        return self.iConfusionMatrix
    
    
    def getDontKnow(self):
        """ Return the number of instances that could not be matched by any rule in the agent: (unclassified instances). """
        return self.iDontKnow
    
    
    def getTotal(self):
        """ Return the number of instances that were attempted to be matched by rules in the agent: (all instances). """
        return self.iTotal
 
 
    def getNoMatch(self):
        """ Returns the fraction of instances which could not be matched by the agent's rule set. """  
        if self.iTotal > 0:
            return self.iDontKnow/float(self.iTotal)
        return 0
    
    
    def getAccuracy(self):
        """ Returns the fraction of instances which were matched and correctly classified by the agent's rule set. """
        acc = 0
        for i in range(self.numClass):
            acc += self.iConfusionMatrix[i][i]
        if self.iTotal > 0:
            return acc/float(self.iTotal)
        return 0
    

    def getScaledAccuracy(self):
        """ Return the scaled accuracy. """
        tmpAcc = self.getAccuracy()
        return pow(tmpAcc,cons.accuracyScaleF)
    
    
    def update(self, classification, instanceClass):
        """ Update the performance objects according to a single instance. """
        if classification == -1:
            self.iDontKnow += 1
        else:
            self.iConfusionMatrix[classification][instanceClass] += 1
        self.iTotal += 1

        
    def reset(self):
        """ Resets the performance objects for a new evaluation cycle. """
        for i in range(self.numClass):
            for j in range(self.numClass):
                self.iConfusionMatrix[i][j] = 0
        self.iDontKnow = 0     
        self.iTotal = 0